/*----------------------------------------------------------------------------
 Description:    Ansteuerung Panasonic Kartenleser ZU-95130TE1
                 Pollin Bestellnummer 720 491 

Chip type        Atmel AVR ATMega32 	(siehe Makefile MCU)
Clock frequency  16 MHz 		(siehe Makefile F_CPU)

For AVR-GCC 4  18. August 2009

                          Copyright:  Werner Boellmann   mailto: werner.boellmann@gmx.net
 Usart.c/.h      Portions Copyright:  Radig Ulrich  mailto: mail@ulrichradig.de
 ChipCard Reader Portions Copyright:  Holger Buss  mailto: buss@mikrocontroller.com

 Author:         Werner Boellmann
 Remarks:        
 known Problems: none
 Version:        15.08.2009

 Dieses Programm ist freie Software. Sie können es unter den Bedingungen der 
 GNU General Public License, wie von der Free Software Foundation veröffentlicht, 
 weitergeben und/oder modifizieren, entweder gemäß Version 2 der Lizenz oder 
 (nach Ihrer Option) jeder späteren Version. 

 Die Veröffentlichung dieses Programms erfolgt in der Hoffnung, 
 daß es Ihnen von Nutzen sein wird, aber OHNE IRGENDEINE GARANTIE, 
 sogar ohne die implizite Garantie der MARKTREIFE oder der VERWENDBARKEIT 
 FÜR EINEN BESTIMMTEN ZWECK. Details finden Sie in der GNU General Public License. 

 Sie sollten eine Kopie der GNU General Public License zusammen mit diesem 
 Programm erhalten haben. 
 Falls nicht, schreiben Sie an die Free Software Foundation, 
 Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA. 
------------------------------------------------------------------------------*/
/*************\
| Just a Demo |
\*************/

/* Standard AVR functions */
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <stdio.h>
#include <string.h>

#include "usart.h"
#include "portsmacro.h"

/* ##### Magnetic Stripe Reader - Pin definition ######### */
#define RDPORT	A	/* Magnetkartenleser belegt Port A */

#define RDT3BIT	2		/* Track 3 Daten */	
#define RCL3BIT (RDT3BIT+1)	/* Track 3 Takt  */  
#define RDT2BIT	4		/* Track 2 Daten */
#define RCL2BIT (RDT2BIT+1)	/* Track 2 Takt  */
#define RDT1BIT	6		/* Track 1 Daten */
#define RCL1BIT (RDT1BIT+1)	/* Track 1 Takt  */

#define CLKMASK  ((1<<RCL1BIT)|(1<<RCL2BIT)|(1<<RCL3BIT)) /* Pin-Bitmaske Taktsignale   */
#define DATAMASK ((1<<RDT1BIT)|(1<<RDT2BIT)|(1<<RDT3BIT)) /* Pin-Bitmaske Daten         */
#define RDMASK   (DATAMASK|CLKMASK)			  /* Magnetkarten Lese-Bitmaske */

/* Track 3 Interface */
#define RDT3P	RDPORT		/* Track 3 Daten PIN/PORT (A/B/C/D) */
#define RDT3B	RDT3BIT		/* Track 3 Daten PIN/PORT Bitnummer */		
#define RCL3P	RDPORT		/* Track 3 Takt PIN/PORT (A/B/C/D)  */
#define RCL3B	RCL3BIT		/* Track 3 Takt PIN/PORT Bitnummer  */

/* Track 2 Interface (Siehe Track 3) */
#define RDT2P	RDPORT
#define RDT2B	RDT2BIT
#define RCL2P	RDPORT
#define RCL2B	RCL2BIT

/* Track 1 Interface (Siehe Track 3) */
#define RDT1P	RDPORT
#define RDT1B	RDT1BIT
#define RCL1P	RDPORT
#define RCL1B	RCL1BIT

/* ############# Card Detect and Card Loaded signals ######### */

#define PH1P	RDPORT
#define PH1B	0
#define PH2P	RDPORT
#define PH2B    1

#define CARD_DETECTED()  	(!PINBIT(PH1P, PH1B))
#define CARD_LOADED()		(!PINBIT(PH2P, PH2B))


/* ############# Eject Selenoid ############# */
#define	SLNDP	C	/* Auswurfmagnet      */
#define	SLNDB	0

/* ####### ChipCard Reader define ########## */

/* Telefon-Chipkartenleser */
#define VCC_BIT		0
#define RESET_BIT	1
#define CLOCK_BIT	2
#define DATA_BIT	3
#define READER_IO	B

#define VCC(v)		do { if (v) PORT(READER_IO) |=  (1<<VCC_BIT); else PORT(READER_IO) &= ~(1<<VCC_BIT); } while(0)
#define RESET(v)	do { if (v) PORT(READER_IO) |=  (1<<RESET_BIT); else PORT(READER_IO) &= ~(1<<RESET_BIT); } while(0)
#define CLOCK(v)    	do { if (v) PORT(READER_IO) |=  (1<<CLOCK_BIT); else PORT(READER_IO) &= ~(1<<CLOCK_BIT); } while(0)
#define DATA_IN		((PIN(READER_IO) & (1<<DATA_BIT)) != 0)
#define DATA_OUT(v)	do { if(v) PORT(READER_IO) |= (1<<DATA_BIT); else PORT(READER_IO) &= ~(1<<DATA_BIT); } while(0)

#define READ  do { PORT(READER_IO) |= (1<<DATA_BIT); DDR(READER_IO) &= ~(1<<DATA_BIT); } while(0)
#define WRITE do { DDR(READER_IO) |=  (1<<DATA_BIT); } while(0)

/* Temporary use */
/* ------------- */
#define LED_INIT	DDR(D) |= ((1<<5)|(1<<6)); 
#define LED1(v)		do { if(v) PORT(D) |= (1<<5); else PORT(D) &= ~(1<<5); } while(0)
#define LED2(v)		do { if(v) PORT(D) |= (1<<6); else PORT(D) &= ~(1<<6); } while(0)


/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

static const char crlf[]    PROGMEM = "\r\n";
static const char dshline[] PROGMEM = "\r\n---------------------\r\n";

/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

void readPhonecard(void);
void readMagstripe_INSERT(void);
/*void readMagstripe_REMOVE(void);*/
void eject(void);

/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

/*
 * Input/Output Ports initialization 
 */
void initMCU(void)
{
	/*
	 * Port A	- Magnetic strip reader
	 */
	DDR(RDPORT) = 0;	/* All IN */
	PORT(RDPORT) = 0;	/* Pull UPs off */

	/*
	 * Port D	- LEDs, USART
	 */
	PORT(D) = 0x00;
	DDR(D)  = 0x00;
	LED1(0);
	LED2(0);
	LED_INIT;

	/*
	 * Port C	- Eject Selenoid
	 */
	SETPORTBIT(SLNDP, SLNDB);
	PORTBITOUT(SLNDP, SLNDB);

	/* 
	 * Port B 	- ChipCard reader 
	 */
	PORT(READER_IO) = (1<<RESET_BIT)|(1<<CLOCK_BIT)|(1<<DATA_BIT);
	DDR(READER_IO)  = (1<<RESET_BIT)|(1<<CLOCK_BIT)|(1<<VCC_BIT);

	/*
	 * Analog Comparator initialization
	 * Analog Comparator: Off
	 * Analog Comparator Input Capture by Timer/Counter 1: Off
	 */
	ACSR = 0x80;
}

int main(void)                              
{
	static const char hello[]         PROGMEM = "Werner's CardReader v0.15 (" __DATE__ ")\r\n";
	static const char removecard[]    PROGMEM = "Please remove card\r\n";
	static const char insertcard[]    PROGMEM = "Please insert card\r\n";

	initMCU();

	/*
	 * UART initialization
	 * Communication Parameters: 8 Data, 1 Stop, No Parity
	 * UART Receiver: On
	 * UART Transmitter: On
	 * UART Baud rate: 38400 bPS
	 * Parity Mode Disabled
	 */
	usart_init(38400);

	sei();

	usart_write_P(hello);

	READ;
	CLOCK(0);	/* Everything is passive */
	RESET(0);
	VCC(0);		/* No Power */
	if( CARD_DETECTED() )
	{
	    LED1(1);
		usart_write_P(removecard);
	
		while( CARD_DETECTED() )
		{
		    if( CARD_LOADED() )
		    {
		    	LED2(1);
		    	LED1(1);
  				eject();
				_delay_ms(150);
		    }
		}
	}
    	LED2(0);
    	LED1(0);
	while (1)
    	{
		READ;
		CLOCK(0);	/* Everything is passive */
		RESET(0);
		VCC(0);		/* No Power */

		_delay_ms(450);		/* For the humans to recognize a delay ;-) */

		while( CARD_DETECTED() )
		{
			if( CARD_LOADED() )
			{
			    	LED2(1);
			    	LED1(1);
	  			eject();
				usart_write_P(removecard);
			}
		    	else LED2(0);
/*			readMagstripe_REMOVE(); */		
		}
	    	LED1(0);
		/* The card slot is empty. Request for a card */
		usart_write_P(insertcard);
		while( !CARD_DETECTED() )
			;		/* Wait for card */
	    	LED1(1);
		readMagstripe_INSERT();

		if( CARD_LOADED() )	/* Fully inserted. Try to read card chip */
		{
			LED2(1);
			readPhonecard();
		}
	}
	return 0;
}

/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *\
 * Card reader - Common subroutines      *
\* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/*
 * Card Eject
 */
void eject(void)
{
	if(CARD_LOADED())
	{
		CLRPORTBIT(SLNDP, SLNDB);
		_delay_ms(70);
		SETPORTBIT(SLNDP, SLNDB);
	}
}

/*
 * Bitreverse a nibble (or byte)
 */
#define SWAP_ONLY_NIBBLE

static unsigned char Mirror(unsigned char input)
{
#ifndef SWAP_ONLY_NIBBLE
	input = ((input & 0x0F)<<4)|((input & 0xF0)>>4);
#endif
	input = ((input & 0x33)<<2)|((input & 0xCC)>>2);
	return ((input & 0x55)<<1)|((input & 0xAA)>>1);
}

/*
 * Count the number of bits set
 */
static unsigned char CntBits(unsigned char input)
{                 
    	unsigned char temp = 0;
     
	while(input)
	{
		if(input & 1) 
			temp++;
		input >>= 1;     
	} 
	return(temp);
}

/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *\
 * ChipCard reader                           *
 * ----------------------------------------- *
 * This code is based on Holger Buss's       *
 * Telephone Card decoder.                   *
 * http://www.mikrocontroller.net/topic/1459 *
\* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

unsigned char memory[33];

/* Clock the Chip */
#define TAKT { CLOCK(1); _delay_us(10); CLOCK(0); _delay_us(10);}

/* Table of available amounts of money stored on the card */  
static const unsigned int tabelle[] PROGMEM = {0,0,0,0,0,0,0,0,5000,6000,1200,600,150,300,0,0};

/*
 * Read a nibble, starting with bit #3
 */
unsigned char read_nibble(void)
{                 
    	unsigned char temp = 0;
    	unsigned char mask = (1 << 3);
 
	while(mask)
	{
		if(DATA_IN) 
			temp |= mask;
		TAKT;      
		mask >>= 1;     
	}
	return(temp);
}

/*
 * Read a European telephone card 
 */
void readPhonecard(void)
{
	VCC(1);			/* Power up chipcard */
	_delay_ms(10);
	CLOCK(1);		/* Passive */
	_delay_ms(10);
	LED2(1);
	RESET(1);
	TAKT;
	RESET(0);
      
	/* Lese Telefonkarte... */
             
	for(unsigned char i = 0; i < 32; i++) 
		memory[i] = read_nibble();
	LED2(0);
	if(((memory[0] & 0x0f) != 0x0f) || ((memory[1] & 0x0f) != 0x0f) || ((memory[2] & 0x0f) != 0x0f) || ((memory[3] & 0x0f) != 0x0f))
	{                 
	 	usart_write_P(dshline);
		usart_write_p("Ident: %1x%1x\r\n", memory[0], memory[1]);
		usart_write_p("Land:  %1x%1x", memory[2], memory[3]);
 
	 	/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
 
	 	usart_write_P(dshline);
		usart_write_p("Kartenwert:  %u\r\n", pgm_read_word(&tabelle[memory[8]]));
        
	 	{
	  		unsigned int value;
	  		value  =  (unsigned int) CntBits(memory[16]) * 4096 + (unsigned int) CntBits(memory[17]) * 4096;
	  		value  += (unsigned int) CntBits(memory[18]) * 512  + (unsigned int) CntBits(memory[19]) * 512;
	  		value  += (unsigned int) CntBits(memory[20]) * 64   + (unsigned int) CntBits(memory[21]) * 64;
	  		value  += (unsigned int) CntBits(memory[22]) * 8    + (unsigned int) CntBits(memory[23]) * 8;
	  		value  += (unsigned int) CntBits(memory[24]) * 1    + (unsigned int) CntBits(memory[25]) * 1;

		  	usart_write_p("Kontostand: %u",value);
	 	}

	 	usart_write_P(dshline);
	 	usart_write_p("Seriennummer: %1x%1x%1x%1x%1x\r\n",
				Mirror(memory[15]),Mirror(memory[14]),
				Mirror(memory[13]),Mirror(memory[12]),
				Mirror(memory[11]));
		usart_write_p("Gueltig bis: %d.%1x",Mirror(memory[10]),Mirror(memory[9]));
	} /* else Keine Telefonkarte */
	usart_write_P(dshline);
}


/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *\
 * Magnetic Stripe Card reader           *
\* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static const char * decodeError(int err)
{
	static const char ok[]      PROGMEM = ".: OK";
	static const char parity[]  PROGMEM = "P: Parity error";
	static const char noiso[]   PROGMEM = "0: Track empty or not ISO formatted";
	static const char noES[]    PROGMEM = "E: Missing End Sentinel. Please try again";
	static const char lrc[]     PROGMEM = "L: LRC Checksum Error";
	static const char unknown[] PROGMEM = "?: Unknown error";

	if(err >= 0)
		return ok;
	switch(err)
	{
	case -1: return parity;
	case -2: return noiso;
	case -3: return noES;
	case -4: return lrc;
	}

	return unknown;
}


/*
 * Decode 5 bit data
 */
int8_t r5decode(uint8_t code)
{
	uint8_t p = 0;

	if(!(CntBits(code) & 1))
		p = 0x80;	/* Parity error	*/

	code &= 0x0F;		/* Mask parity bit */ 

	if(code <= 9)
		return ((code + '0') | p);
	if(code == 0b00001011)
		return ('s' | p);	/* Start Sentinel */
	if(code == 0b00001101)
	{
		return ('^' | p);	/* Separator */
	}
	if(code == 0b00001111)
		return ('e' | p);	/* End Sentinel */
	return ('?' | p);		/* unknown */
}

/*
 * Find 5 bit start sentinel and decode
 * Returns: Bit number where Start Sentinel was found negative value on error
 * -1 : Parity Error
 * -2 : Track Empty or not ISO formatted
 * -3 : Start Sentinel found, but not End Sentinel
 * -4 : LRC Error
 */
static int findSSpattern5(uint8_t *p, int len)
{
	#define pattern5  ((uint8_t)0b00001011)
	#define SH	5
	#define SHMASK  ((1<<SH)-1)
	#define SHMASK_1  ((1<<(SH-1))-1)

	int i          = 0;	/* Bits processed     */
	uint8_t mask   = 1;     /* Bitmask in byte    */ 
	uint8_t fp     = 0;     /* Temp store         */
	uint8_t SEflag = 0;	/* End Sentinel found */
	uint8_t LRC    = 0;

	while(i < len)		/* Search all... */
	{
		fp >>= 1;		/* shift search window */
		if(p[i/8] & mask)
			fp |= 0x80;	/* Window vaild */
		if(!(mask <<= 1)) 	/* Mask rollover */
			mask = 1;	/* Reset start value */
		i++;			/* Bit moved to search window */
		if((fp & SHMASK) == pattern5)	/* Start Sentinel found */
		{
			uint8_t bitCnt;		/* to count 5 bits */
			char    c; 		/* Decoded content */	
			int offset = i-SH;	/* Function return value */

			do
			{
				LRC ^= (fp & SHMASK_1);		/* update LRC */
				c = r5decode(fp & SHMASK);  	/* Decode value */
				if(c & 0x80 && !SEflag)  	/* Parity Error */
					LRC |= 0x80;
				c &= 0x7F;
				if(SEflag)  	/* End */
				{
					if(LRC & 0x80)		/* Parity error */
						return -1;
					LRC &= SHMASK_1;
					if(LRC != 0)
						return -4;	/* LRC Error */
					return offset;		/* End Sentinel Seen */
				}
				usart_write_char(c);
				if(c == 'e')			/* Remember: End Sentinel seen */
					SEflag++;
				if(c == '^' || c == 'e')	/* Remember: End Sentinel seen */
					usart_write_P(crlf);
				/* Grab the next 5 bits into the window and decode */
				for(bitCnt = 0; (i < len) && (bitCnt < SH); bitCnt++)
				{
					fp >>= 1;		/* Shift decode window */
					if(p[i/8] & mask)	/* Update value */
						fp |= 0x80;
					if(!(mask <<= 1)) 	/* Mask rollover */
						mask = 1;	/* Reset to initial value */
					i++;			/* Bit processed. Next! */
				}
			} while(i < len);			/* Until done or Parity error */
			if(LRC & 0x80)
				return -1;
			if(!SEflag) 	  /* End Sentinel not seen -> Error */
				return -3;
			return offset;	  /* End Sentinel Seen, so Partity Error is not valid */
		}
	}
	return -2;	/* Track empty or not ISO formatted */
	#undef pattern5
	#undef SHMASK_1
	#undef SHMASK
	#undef SH
}

/*
 * Decode 7 bit data (Track 1)
 */
int8_t r7decode(uint8_t p)
{
	uint8_t q = 0;

	if((CntBits(p) & 1) == 0)
		q = 0x80;	/* Parity error */

	p &= 0x3F;		/* Mask parity bit */

	if(p <  0b010000)
	{
		if(p == 0b000000)
			return (' ' | q);
		if(p == 0b000100)
			return ('$' | q);
		if(p == 0b000101)
			return ('s' | q);	/* Start Sentinel */
		if(p == 0b001101)
			return ('-' | q);
		if(p == 0b001110)
			return ('.' | q);
		if(p == 0b001111)
			return ('/' | q);
		return ('?' | q);
	}
	if(p <= 0b011001)
		return ((p - 0b010000 + '0') | q);
	if(p == 0b011111)
		return ('e' | q);	/* End Sentinel */
	if(p == 0b100000)
		return ('?' | q);	/* Undefined */
	if(p <  0b111011)
		return ((p - 0b100001 + 'A') | q);
	if(p == 0b111011)
		return ('[' | q);
	if(p == 0b111100)
		return ('/' | q);
	if(p == 0b111101)
		return (']' | q);
	if(p == 0b111110)
		return ('^' | q);	/* Separator */
	return ('?' | q);		/* unknown */
}

/*
 * Find 7 bit start sentinel and decode (Track 1)
 * Returns: Bit number where Start Sentinel found or negative value on error.
 * -1 : Parity Error
 * -2 : Track Empty or not ISO formatted
 * -3 : Start Sentinel found, but not End Sentinel
 * -4 : LRC Error
 */
static int findSSpattern7(uint8_t *p, int len)
{
	#define pattern7  ((uint8_t)0b01000101)
	#define SH	7
	#define SHMASK  ((1<<SH)-1)
	#define SHMASK_1 ((1<<(SH-1))-1)

	int     i      = 0;	/* Bits processed         */
	uint8_t mask   = 1;     /* Bitmask in RAM byte    */
	uint8_t fp     = 0;	/* Data Window            */
	uint8_t SEflag = 0;     /* End Sentinel seen flag */
	uint8_t LRC    = 0;

	while(i < len)
	{
		fp >>= 1;		/* move window */
		if(p[i/8] & mask)	/* Bit to move from RAM to Window */ 
			fp |= 0x80;
		if(!(mask <<= 1)) 	/* Mask rollover detected */
			mask = 1;	/* Reset to start value */
		i++;			/* Bit "Processed" */
		if((fp & SHMASK) == pattern7)	/* Start Sentinel ? */
		{
			uint8_t bitCnt;		/* Count seve bits per symbol */
			char    c; 		/* Decoded symbol */	
			int offset = i-SH;	/* Remember return value. */ 

			do
			{
				LRC ^= fp & SHMASK_1;
				c = r7decode(fp & SHMASK); 	/* Decode */
				if(c & 0x80)
					LRC |= 0x80;		/* Parity Error */
				c &= 0x7F;
				if(SEflag)		 	/* End */
				{
					if(LRC & 0x80) 		/* Parity Error */
						return -1;
					if(LRC & SHMASK_1)
						return -4;	/* LRC Error */
					return offset;
				}
				usart_write_char(c);
				if(c == 'e') 			/* Mark "End Sentinel seen" */
					SEflag++;
				if(c == '^' || c == 'e') 	/* Record separator seen */
					usart_write_P(crlf);

				/* Move the read window seven bit further */
				for(bitCnt = 0; (i < len) && (bitCnt < SH); bitCnt++)
				{
					fp >>= 1;		/* Move window */
					if(p[i/8] & mask)	/* Tranfer bit */
						fp |= 0x80;
					if(!(mask <<= 1)) 	/* mask overflow */
						mask = 1;	/* Reset to start value */
					i++;
				}
			} while(i < len);
			if(LRC & 0x80)
				return -1;
			if(!SEflag) 	/* Error if End Sentinel not seen */
				return -3;
			return offset;
		}
	}
	return -2;
	#undef pattern7
	#undef SHMASK_1
	#undef SHMASK
	#undef SH
}

/* Bit-buffer for tracks */
uint8_t iso1[100];  /*  79 characters (7 bit code), 210 BPI + spare */
uint8_t iso2[50];   /*  40 characters (5 bit code),  75 BPI + spare */
uint8_t iso3[130];  /* 107 characters (5 bit code), 210 BPI + spare */

/*
 * Read while insert (backwards)
 */
void readMagstripe_INSERT(void)
{
	static const char    info[]     PROGMEM = "\r\n%d. track: %d bits read\r\n";
	static const uint8_t bitsmask[] PROGMEM = { 1,2,4,8,16,32,64,128 }; /* Use: pgm_read_byte(&bitsmask[iX % 8]); */
	uint8_t	state, newstate;/* Current and last clock signal levels */
	int i1, i2, i3;		/* Bit counter for the three tracks */

	state = ~PIN(RDPORT) & CLKMASK;		/* read the current state of the clock signal bits */
	
	/*
	 * Because the data is read backwards, i also decided to store it backwards.
	 * Place index behind the last byte of the corresponding buffer
	 */
	i1 = sizeof(iso1) * 8;
	i2 = sizeof(iso2) * 8;
	i3 = sizeof(iso3) * 8;

	/* Clear the buffers */
	memset(iso1, 0, sizeof(iso1));
	memset(iso2, 0, sizeof(iso2));
	memset(iso3, 0, sizeof(iso3));

	while( CARD_DETECTED() && !CARD_LOADED() )  /* while Card detect but not loaded yet */
	{
		/* Detect a falling edge */
		newstate = ~PIN(RDPORT) & CLKMASK;  /* current state of the clock bits */
		state ^= newstate;		    /* changed clock bits are 1 */
		state &= newstate;		    /* clock bits that have seen a falling edge are 1 */
		if(state)			    /* any clock signal changed to 0 ? */
		{	
			uint8_t data = ~PIN(RDPORT);	/* read the data bits */
			
			/* Falling edge at "Track 1" clock ? */
			if(state & (1<<RCL1BIT))
			{
			    if (i1 >= 0)		/* Any space left in buffer */
			    {
				if(data & (1<<RDT1BIT))	/* Is it "1" bit ? */
					iso1[i1 / 8] |=  pgm_read_byte(&bitsmask[i1 % 8]);
				i1--;
			    }
			}

			/* Falling edge at "Track 2" clock ? */
			if(state & (1<<RCL2BIT))
			{
			    if (i2 >= 0)	
			    {
				if(data & (1<<RDT2BIT))
					iso2[i2 / 8] |=  pgm_read_byte(&bitsmask[i2 % 8]);
				i2--;
			    }
			}

			/* Falling edge at "Track 3" clock ? */
			if(state & (1<<RCL3BIT))
			{
			    if (i3 >= 0)
			    {
				if(data & (1<<RDT3BIT))
					iso3[i3 / 8] |=  pgm_read_byte(&bitsmask[i3 % 8]);
				i3--;
			    }
			}
		}
		/* Remember processed clock state */
		state = newstate;
	}

	/* 
	 * Decode and transfer data to terminal
	 */

	static const char result[] PROGMEM = "%S (%d).\r\n";

	/* Track 1 */
	usart_write_P(info, 1, sizeof(iso1) * 8 - i1);
	i1 = findSSpattern7(iso1, sizeof(iso1) * 8);
	usart_write_P(result, decodeError(i1), i1);

	/* Track 2 */
	usart_write_P(info, 2, sizeof(iso2) * 8 - i2);
	i2 = findSSpattern5(iso2, sizeof(iso2) * 8);
	usart_write_P(result, decodeError(i2), i2);

	/* Track 3 */
	usart_write_P(info, 3, sizeof(iso3) * 8 - i3);
	i3 = findSSpattern5(iso3, sizeof(iso3) * 8);
	usart_write_P(result, decodeError(i3), i3);
}

#if 0
/*
 * Read while removing (from the beginning)
 */
void readMagstripe_REMOVE(void)
{
	static const char info[]        PROGMEM = "\r\n%d. track: %d bits read\r\n";
	static const uint8_t bitsmask[] PROGMEM = { 1,2,4,8,16,32,64,128 }; /* Use: pgm_read_byte(&bitsmask[iX % 8]);n */
	uint8_t	state, newstate;/* Current and last clock signal levels */
	int i1, i2, i3;		/* Bit counter for the three tracks */

	state = ~PIN(RDPORT) & CLKMASK;		/* read the current state of the clock signal bits */
	
	/*
	 * Place index at the first bit of the corresponding buffer
	 */
	i1 = i2 = i3 = 0;

	/* Clear the buffers */
	memset(iso1, 0, sizeof(iso1));
	memset(iso2, 0, sizeof(iso2));
	memset(iso3, 0, sizeof(iso3));

	while( CARD_DETECTED() && !CARD_LOADED() )  /* while Card detect but not loaded */
	{
		/* Detect a falling edge */
		newstate = ~PIN(RDPORT) & CLKMASK;  /* current state of the clock bits */
		state ^= newstate;		    /* changed clock bits are 1 */
		state &= newstate;		    /* clock bits that have seen a falling edge are 1 */
		if(state)			    /* any clock signal changed to 0 ? */
		{	
			uint8_t data = ~PIN(RDPORT);	/* read the data bits */
			
			/* Falling edge at "Track 1" clock ? */
			if((state & (1<<RCL1BIT)))
			{
			    if (i1 < (sizeof(iso1)*8))		/* Any space left in buffer */
			    {
				if(data & (1<<RDT1BIT))	/* Is it "1" bit ? */
					iso1[i1 / 8] |=  pgm_read_byte(&bitsmask[i1 % 8]);
			    }
			    i1++;
			}

			/* Falling edge at "Track 2" clock ? */
			if(state & (1<<RCL2BIT))
			{
			    if (i2 < (sizeof(iso2)*8))	
			    {
				if(data & (1<<RDT2BIT))
					iso2[i2 / 8] |=  pgm_read_byte(&bitsmask[i2 % 8]);
			    }
			    i2++;
			}

			/* Falling edge at "Track 3" clock ? */
			if(state & (1<<RCL3BIT))
			{
			    if (i3 < (sizeof(iso3)*8))
			    {
				if(data & (1<<RDT3BIT))
					iso3[i3 / 8] |=  pgm_read_byte(&bitsmask[i3 % 8]);
			    }
			    i3++;
			}
		}
		/* Remember processed clock state */
		state = newstate;
	}

	if(CARD_LOADED())
		return; /* ERROR, not removed */
	/* 
	 * Decode and transfer data to terminal
	 */

	static const char result[] PROGMEM = "%S (%d).\r\n";

	/* Track 1 */
	usart_write_P(info, 1, sizeof(iso1) * 8 - i1);
	i1 = findSSpattern7(iso1, sizeof(iso1) * 8);
	usart_write_P(result, decodeError(i1), i1);

	/* Track 2 */
	usart_write_P(info, 2, sizeof(iso2) * 8 - i2);
	i2 = findSSpattern5(iso2, sizeof(iso2) * 8);
	usart_write_P(result, decodeError(i2), i2);

	/* Track 3 */
	usart_write_P(info, 3, sizeof(iso3) * 8 - i3);
	i3 = findSSpattern5(iso3, sizeof(iso3) * 8);
	usart_write_P(result, decodeError(i3), i3);

}
#endif

